/* $XConsortium: keys.h,v 1.3 93/10/28 15:27:11 gildea Exp $ */

/*

Copyright 1989-1991, Bitstream Inc., Cambridge, MA.
You are hereby granted permission under all Bitstream propriety rights to
use, copy, modify, sublicense, sell, and redistribute the Bitstream Speedo
software and the Bitstream Charter outline font for any purpose and without
restrictions; provided, that this notice is left intact on all copies of such
software or font and that Bitstream's trademark is acknowledged as shown below
on all unmodified copies of such font.

BITSTREAM CHARTER is a registered trademark of Bitstream Inc.


BITSTREAM INC. DISCLAIMS ANY AND ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING
WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE.  BITSTREAM SHALL NOT BE LIABLE FOR ANY DIRECT OR INDIRECT
DAMAGES, INCLUDING BUT NOT LIMITED TO LOST PROFITS, LOST DATA, OR ANY OTHER
INCIDENTAL OR CONSEQUENTIAL DAMAGES, ARISING OUT OF OR IN ANY WAY CONNECTED
WITH THE SPEEDO SOFTWARE OR THE BITSTREAM CHARTER OUTLINE FONT.

*/


/***** DECRYPTION KEY CONSTANTS (PC Platform) *****/

#define CUS0  432                  /* Customer number */

#define KEY0    0                  /* Decryption key 0 */
#define KEY1   72                  /* Decryption key 1 */
#define KEY2  123                  /* Decryption key 2 */
#define KEY3    1                  /* Decryption key 3 */
#define KEY4  222                  /* Decryption key 4 */
#define KEY5  194                  /* Decryption key 5 */
#define KEY6  113                  /* Decryption key 6 */
#define KEY7  119                  /* Decryption key 7 */
#define KEY8   52                  /* Decryption key 8 */

/***** DECRYPTION KEY CONSTANTS (Sample) *****/

#define XSAMPLEFONTS

#define XCUS0    0                 /* Customer number */

#define XKEY0    0                  /* Decryption key 0 */
#define XKEY1    0                  /* Decryption key 1 */
#define XKEY2    0                  /* Decryption key 2 */
#define XKEY3    0                  /* Decryption key 3 */
#define XKEY4    0                  /* Decryption key 4 */
#define XKEY5    0                  /* Decryption key 5 */
#define XKEY6    0                  /* Decryption key 6 */
#define XKEY7    0                  /* Decryption key 7 */
#define XKEY8    0                  /* Decryption key 8 */


