/* REXX */
/* $XFree86: xc/config/util/xmake.cmd,v 3.1 1996/03/10 11:50:47 dawes Exp $ */
'@echo off'
PARSE ARG a1 a2 a3 a4 a5 a6 a7 a8 a9 a10
x11make.exe 'MAKE=x11make.exe SHELL=' a1 a2 a3 a4 a5 a6 a7 a8 a9 a10
